unit dnDBPxM;

(*****************************************************************************
Copyright  1997 Dieter Nth
version    : 1.02
date       : 19.07.1997
description:

  Component TPxTable: only one new public property (read only): PxMasks,
    i.e. Paradox-masks defined at table-level. If a TDBPxEdit is linked
    to a TPxTable, then TDBPxEdit loads the PxMask from it.

  Component TDBPxEdit, new properties:
    PxMask: 1. if DataSource is linked to a TPxTable, loads PxMask from TPxTable,
            2. if not, loads/stores PxMask from/to a DataSet linked to PxMaskSource
    PxMaskSource: DataSet to load PxMask from, this is done via
      D1: TTable.FindKey --> Index must be properly set
      D2: TDataset.locate --> Table must have Fields 'FieldName' and 'PdoxMask'
          (I use locate instead of lookup for a CaseInsensitive search of Fieldname)
      if PxMaskSource = nil, it works like TPxMaskEdit

  a part of the TDBPxEdit-class is based upon code from TDBEdit

Bugs fixed:
1.02 :
  stupid bug: In D2 method LoadPxMasks didn't work, because I changed code a bit
  (working with D1 and testing with D1) defined STableName as
  array[0..SizeOf(TFileName) - 1] of Char. In D1 TFileName is defined as
  String[79], but in D2 it's just string (= AnsiString = pointer),
  so sizeOf(TFileName) returns 4 instead, causing every filename truncated to
  first 3 chars. sorry.

******************************************************************************)

interface

uses
  {$IFDEF Ver80} WinTypes, WinProcs, {$ELSE} Windows, {$ENDIF}
  SysUtils, Messages, Classes, Controls, Forms, Graphics, Menus, StdCtrls,
  DB, DBTables, DBCtrls, Buttons, DBITypes,
  dnPxMask;

const
  {only D2: Field used in locate}
  cDefaultNameField = 'FieldName';
  {Field to retrieve Paradoxmask from}
  cDefaultNameMask = 'PdoxMask';
  {This error is raised only in Designmode, if you try to set a Mask defined in
   a Paradoxtable}
  cMaskDefinedOnTableLevel =
    'Inputmask is defined on Tablelevel (Paradox) and can''t be changed';
{    'Eingabemaske ist auf Tabellen-Ebene (Paradox) festgelegt und kann nicht gendert werden'}

type
  TPxTable = class(TTable)
  private
    { Private-Deklarationen }
    FPxMasks : TStrings;
    function GetPxMask(const Index : string) : string;
    procedure LoadPxMasks;
    procedure ClearPxMasks;
  protected
    { Protected-Deklarationen }
    procedure CloseCursor; override;
    procedure OpenCursor; override;
  public
    { Public-Deklarationen }
    destructor Destroy; override;
    property PxMasks[const index : string] : string read GetPxMask;
  end;

  TDBPxEdit = class(TCustomPxMaskEdit)
  private
    FMaskLink: TFieldDataLink;
    FDataLink: TFieldDataLink;
    FCanvas: TControlCanvas;
    FAlignment: TAlignment;
    FFocused: Boolean;
    FStored: Boolean;

    {$IFDEF Ver80}
    FTextMargin: Integer;
    procedure CalcTextMargin;
    {$ENDIF}

    procedure DataChange(Sender: TObject);
    procedure EditingChange(Sender: TObject);
    procedure ActiveChange(Sender: TObject);
    function GetDataField: string;
    function GetDataSource: TDataSource;
    function GetMaskSource: TDataSource;
    procedure SetMaskSource(Value: TDataSource);
    function GetField: TField;
    function GetReadOnly: Boolean;
    function GetPxMask : string;

    {$IFNDEF Ver80}
    function GetTextMargins: TPoint;
    {$ENDIF}

    function GetIndex : string;
    procedure SetDataField(const Value: string);
    procedure SetDataSource(Value: TDataSource);
    procedure SetFocused(Value: Boolean);
    procedure SetReadOnly(Value: Boolean);
    procedure SetPxMask(const Value: string);
    procedure UpdateData(Sender: TObject);
    procedure WMCut(var Message: TMessage); message WM_CUT;
    procedure WMPaste(var Message: TMessage); message WM_PASTE;
    procedure CMEnter(var Message: TCMEnter); message CM_ENTER;
    procedure CMExit(var Message: TCMExit); message CM_EXIT;
    procedure WMPaint(var Message: TWMPaint); message WM_PAINT;

    {$IFDEF Ver80}
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    {$ELSE}
    procedure CMGetDataLink(var Message: TMessage); message CM_GETDATALINK;
    {$ENDIF}

  protected
    procedure Change; override;
    function EditCanModify: Boolean; override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: Char); override;
    procedure Notification(AComponent: TComponent;
      Operation: TOperation); override;
    procedure Reset; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property Field: TField read GetField;
    property Modified;
  published
    property AutoSelect;
    property AutoSize;
    property BorderStyle;
    property CharCase;
    property Color;
    property Ctl3D;
    property DataField: string read GetDataField write SetDataField;
    property DataSource: TDataSource read GetDataSource write SetDataSource;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Font;
    property MaxLength;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PasswordChar;
    property PopupMenu;
    property PxMask read GetPxMask write SetPxMask stored FStored;
    property PxMaskSource: TDataSource read GetMaskSource write SetMaskSource;
    property ReadOnly: Boolean read GetReadOnly write SetReadOnly default False;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnChange;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;

    {$IFNDEF Ver80}
    property OnStartDrag;
    {$ENDIF}
  end;

implementation

uses
  DBIErrs, DBIProcs, DBConsts, Dialogs;

{ TDBPxEdit }

constructor TDBPxEdit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  inherited ReadOnly := True;
  FDataLink := TFieldDataLink.Create;
  FDataLink.Control := Self;
  FDataLink.OnActiveChange := ActiveChange;
  FDataLink.OnDataChange := DataChange;
  FDataLink.OnEditingChange := EditingChange;
  FDataLink.OnUpdateData := UpdateData;
  FMaskLink := TFieldDataLink.Create;
  {$IFDEF Ver80}
  CalcTextMargin;
  {$ELSE}
  ControlStyle := ControlStyle + [csReplicatable];
  {$ENDIF}
end;

destructor TDBPxEdit.Destroy;
begin
  FDataLink.Free;
  FDataLink := nil;
  FMaskLink.Free;
  FMaskLink := nil;
  FCanvas.Free;
  inherited Destroy;
end;

procedure TDBPxEdit.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (FDataLink <> nil) and
    (AComponent = DataSource) then DataSource := nil;
end;

procedure TDBPxEdit.KeyDown(var Key: Word; Shift: TShiftState);
begin
  inherited KeyDown(Key, Shift);
  if (Key = VK_DELETE) or ((Key = VK_INSERT) and (ssShift in Shift)) then
    FDataLink.Edit;
end;

procedure TDBPxEdit.KeyPress(var Key: Char);
begin
  if (Key in [#32..#255]) and (FDataLink.Field <> nil) and
    not FDataLink.Field.IsValidChar(Key) then
  begin
    MessageBeep(0);
    Key := #0;
  end;
  case Key of
    ^H, ^V, ^X, #32..#255 : FDataLink.Edit;
  end;
  inherited KeyPress(Key);
end;

function TDBPxEdit.EditCanModify: Boolean;
begin
  Result := FDataLink.Edit;
end;

procedure TDBPxEdit.Reset;
begin
  FDataLink.Reset;
  SelectAll;
end;

procedure TDBPxEdit.SetFocused(Value: Boolean);
begin
  if FFocused <> Value then
  begin
    FFocused := Value;
    if (FAlignment <> taLeftJustify) and not IsMasked then Invalidate;
    FDataLink.Reset;
  end;
end;

procedure TDBPxEdit.Change;
begin
  FDataLink.Modified;
  inherited Change;
end;

function TDBPxEdit.GetDataSource: TDataSource;
begin
  Result := FDataLink.DataSource;
end;

procedure TDBPxEdit.SetDataSource(Value: TDataSource);
begin
  FDataLink.DataSource := Value;
  {$IFNDEF Ver80}
  if Value <> nil then Value.FreeNotification(Self);
  {$ENDIF}
end;

function TDBPxEdit.GetMaskSource: TDataSource;
begin
  Result := FMaskLink.DataSource;
end;

procedure TDBPxEdit.SetMaskSource(Value: TDataSource);
begin
  FMaskLink.DataSource := Value;
  {$IFNDEF Ver80}
  if Value <> nil then Value.FreeNotification(Self);
  {$ENDIF}
end;

function TDBPxEdit.GetDataField: string;
begin
  Result := FDataLink.FieldName;
end;

procedure TDBPxEdit.SetDataField(const Value: string);
begin
  FDataLink.FieldName := Value;
  ActiveChange(self);
end;

function TDBPxEdit.GetReadOnly: Boolean;
begin
  Result := FDataLink.ReadOnly;
end;

procedure TDBPxEdit.SetReadOnly(Value: Boolean);
begin
  FDataLink.ReadOnly := Value;
end;

function TDBPxEdit.GetField: TField;
begin
  Result := FDataLink.Field;
end;

function TDBPxEdit.GetIndex : string;
var
  ds : TDataSet;
begin
  result := '';
  ds := FDataLink.DataSet;
  if ds is TTable then
    result := ChangeFileExt(TTable(ds).TableName, '')
  else
  if ds is TStoredProc then
    result := TStoredProc(ds).StoredProcName
  else
    result := ds.name;
  if (result <> '') and (DataField <> '') then
    result := Format('%s.%s', [result, DataField]);
end;

function TDBPxEdit.GetPxMask : string;
var
  ds : TDataSet;
begin
  if (DataField <> '') then
  begin
    result := '';
    ds := FDataLink.DataSet;
    if ds is TPxTable then
      result := TPxTable(ds).PxMasks[DataField];
    if (result = '') then
    begin
      ds := FMaskLink.DataSet;
      if assigned(ds) and (DataField <> '') and ds.active then
      begin
        {$IFDEF Ver80}
        if (ds is TTable) and TTable(ds).FindKey([GetIndex]) then
        {$ELSE}
        if ds.Locate(cDefaultNameField, GetIndex, [loCaseInsensitive]) then
        {$ENDIF}
          result := ds.FieldByName(cDefaultNameMask).asString;
      end;
    end;
    FStored := result = '';
    if not FStored then inherited PxMask := result;
  end;
  result := inherited PxMask;
end;

procedure TDBPxEdit.SetPxMask(const Value: string);
var
  ds : TDataSet;
  index : string;
begin
  if Value <> inherited PxMask then
  begin
    ds := FDataLink.DataSet;
    if (ds is TPxTable) and (TPxTable(ds).PxMasks[DataField] <> '') then
      if (csDesigning in ComponentState) then
        raise exception.create(cMaskDefinedOnTableLevel)
      else
        exit;
    inherited PxMask := Value;
    FStored := true;
    ds := FMaskLink.DataSet;
    if assigned(ds) and (DataField <> '') then
    begin
      if not ds.active then DBError(SDataSetClosed);
      Index := GetIndex;
      {$IFDEF Ver80}
      if (ds is TTable) and TTable(ds).FindKey([GetIndex]) then
      {$ELSE}
      if ds.Locate(cDefaultNameField, Index, [loCaseInsensitive]) then
      {$ENDIF}
        ds.edit
      else
        ds.insert;
      ds.FieldByName(cDefaultNameMask).asString := inherited PxMask;
      ds.FieldByName(cDefaultNameField).asString := Index;
      ds.post;
      FStored := false;
    end;
  end;
end;

procedure TDBPxEdit.DataChange(Sender: TObject);
begin
  if FDataLink.Field <> nil then
  begin
    FAlignment := FDataLink.Field.Alignment;
    if FDataLink.Field.DataType = ftString then
      MaxLength := FDataLink.Field.Size else
      MaxLength := 0;
    if FFocused and FDataLink.CanModify then
      Text := FDataLink.Field.Text
    else
      Text := FDataLink.Field.DisplayText;
  end else
  begin
    FAlignment := taLeftJustify;
    MaxLength := 0;
    if csDesigning in ComponentState then
      Text := Name else
      Text := '';
  end;
end;

procedure TDBPxEdit.ActiveChange(Sender: TObject);
begin
  if not FStored then inherited PxMask := '';
  if FDataLink.active then GetPxMask;
end;

procedure TDBPxEdit.EditingChange(Sender: TObject);
begin
  inherited ReadOnly := not FDataLink.Editing;
  if FDataLink.active and FDataLink.Editing then GetPxMask;
end;

procedure TDBPxEdit.UpdateData(Sender: TObject);
begin
  ValidateEdit;
  FDataLink.Field.Text := Text;
end;

procedure TDBPxEdit.WMPaste(var Message: TMessage);
begin
  FDataLink.Edit;
  inherited;
end;

procedure TDBPxEdit.WMCut(var Message: TMessage);
begin
  FDataLink.Edit;
  inherited;
end;

procedure TDBPxEdit.CMEnter(var Message: TCMEnter);
begin
  SetFocused(True);
  inherited;
end;

procedure TDBPxEdit.CMExit(var Message: TCMExit);
begin
  try
    FDataLink.UpdateRecord;
  except
    SelectAll;
    SetFocus;
    raise;
  end;
  SetFocused(False);
  SetCursor(0);
  DoExit;
end;

{$IFDEF Ver80}
procedure TDBPxEdit.WMPaint(var Message: TWMPaint);
var
  Width, Indent, Left, I: Integer;
  R: TRect;
  DC: HDC;
  PS: TPaintStruct;
  S: string;
begin
  if (FAlignment = taLeftJustify) or FFocused then
  begin
    inherited;
    Exit;
  end;
{ Since edit controls do not handle justification unless multi-line (and
  then only poorly) we will draw right and center justify manually unless
  the edit has the focus. }
  if FCanvas = nil then
  begin
    FCanvas := TControlCanvas.Create;
    FCanvas.Control := Self;
  end;
  DC := Message.DC;
  if DC = 0 then DC := BeginPaint(Handle, PS);
  FCanvas.Handle := DC;
  try
    FCanvas.Font := Font;
    with FCanvas do
    begin
      R := ClientRect;
      if (BorderStyle = bsSingle) then
      begin
        Brush.Color := clWindowFrame;
        FrameRect(R);
        InflateRect(R, -1, -1);
      end;
      Brush.Color := Color;
      S := Text;
      if PasswordChar <> #0 then
      begin
        for I := 1 to Length(S) do
          S[I] := PasswordChar;
      end;
      Width := TextWidth(S);
      if BorderStyle = bsNone then Indent := 0 else Indent := FTextMargin;
      if FAlignment = taRightJustify then
        Left := R.Right - Width - Indent else
        Left := (R.Left + R.Right - Width) div 2;
      TextRect(R, Left, Indent, S);
    end;
  finally
    FCanvas.Handle := 0;
    if Message.DC = 0 then EndPaint(Handle, PS);
  end;
end;

procedure TDBPxEdit.CMFontChanged(var Message: TMessage);
begin
  inherited;
  CalcTextMargin;
end;

procedure TDBPxEdit.CalcTextMargin;
var
  DC: HDC;
  SaveFont: HFont;
  I: Integer;
  SysMetrics, Metrics: TTextMetric;
begin
  DC := GetDC(0);
  GetTextMetrics(DC, SysMetrics);
  SaveFont := SelectObject(DC, Font.Handle);
  GetTextMetrics(DC, Metrics);
  SelectObject(DC, SaveFont);
  ReleaseDC(0, DC);
  I := SysMetrics.tmHeight;
  if I > Metrics.tmHeight then I := Metrics.tmHeight;
  FTextMargin := I div 4;
end;

{$ELSE}

procedure TDBPxEdit.WMPaint(var Message: TWMPaint);
var
  Left: Integer;
  Margins: TPoint;
  R: TRect;
  DC: HDC;
  PS: TPaintStruct;
  S: string;
begin
  if ((FAlignment = taLeftJustify) or FFocused) and
    not (csPaintCopy in ControlState) then
  begin
    inherited;
    Exit;
  end;
{ Since edit controls do not handle justification unless multi-line (and
  then only poorly) we will draw right and center justify manually unless
  the edit has the focus. }
  if FCanvas = nil then
  begin
    FCanvas := TControlCanvas.Create;
    FCanvas.Control := Self;
  end;
  DC := Message.DC;
  if DC = 0 then DC := BeginPaint(Handle, PS);
  FCanvas.Handle := DC;
  try
    FCanvas.Font := Font;
    with FCanvas do
    begin
      R := ClientRect;
      if not (NewStyleControls and Ctl3D) and (BorderStyle = bsSingle) then
      begin
        Brush.Color := clWindowFrame;
        FrameRect(R);
        InflateRect(R, -1, -1);
      end;
      Brush.Color := Color;
      if (csPaintCopy in ControlState) and (FDataLink.Field <> nil) then
      begin
        S := FDataLink.Field.DisplayText;
        case CharCase of
          ecUpperCase: S := AnsiUpperCase(S);
          ecLowerCase: S := AnsiLowerCase(S);
        end;
      end else
        S := Text;
      if PasswordChar <> #0 then FillChar(S[1], Length(S), PasswordChar);
      Margins := GetTextMargins;
      case FAlignment of
        taLeftJustify: Left := Margins.X;
        taRightJustify: Left := ClientWidth - TextWidth(S) - Margins.X - 1;
      else
        Left := (ClientWidth - TextWidth(S)) div 2;
      end;
      TextRect(R, Left, Margins.Y, S);
    end;
  finally
    FCanvas.Handle := 0;
    if Message.DC = 0 then EndPaint(Handle, PS);
  end;
end;

procedure TDBPxEdit.CMGetDataLink(var Message: TMessage);
begin
  Message.Result := Integer(FDataLink);
end;

function TDBPxEdit.GetTextMargins: TPoint;
var
  DC: HDC;
  SaveFont: HFont;
  I: Integer;
  SysMetrics, Metrics: TTextMetric;
begin
  if NewStyleControls then
  begin
    if BorderStyle = bsNone then I := 0 else
      if Ctl3D then I := 1 else I := 2;
    Result.X := SendMessage(Handle, EM_GETMARGINS, 0, 0) and $0000FFFF + I;
    Result.Y := I;
  end else
  begin
    if BorderStyle = bsNone then I := 0 else
    begin
      DC := GetDC(0);
      GetTextMetrics(DC, SysMetrics);
      SaveFont := SelectObject(DC, Font.Handle);
      GetTextMetrics(DC, Metrics);
      SelectObject(DC, SaveFont);
      ReleaseDC(0, DC);
      I := SysMetrics.tmHeight;
      if I > Metrics.tmHeight then I := Metrics.tmHeight;
      I := I div 4;
    end;
    Result.X := I;
    Result.Y := I;
  end;
end;
{$ENDIF}


{ TPxTable }

destructor TPxTable.Destroy;
begin
  ClearPxMasks;
  inherited destroy;
end;

procedure TPxTable.OpenCursor;
begin
  inherited OpenCursor;
  LoadPxMasks;
end;

procedure TPxTable.CloseCursor;
begin
  ClearPxMasks;
  inherited CloseCursor;
end;

function TPxTable.GetPxMask(const Index : string) : string;
begin
  if assigned(FPxMasks) then
    result := FPxMasks.Values[index]
  else
    result := '';
end;

procedure TPxTable.ClearPxMasks;
begin
  FPxMasks.free;
  FPxMasks := nil;
end;

procedure TPxTable.LoadPxMasks;
var
  VCursor: HDBICur;
  STableName: DBIPath;
  ValCheckDesc: VCHKDesc;
  s : string;
begin
  if FPxMasks = nil then
    FPxMasks := TStringList.create
  else
    FPxMasks.clear;
  if Handle = nil then exit;
  AnsiToNative(DBLocale, TableName, STableName, SizeOf(STableName) - 1);
  if DbiOpenVChkList(DBHandle, STableName, szParadox, VCursor) = 0 then
  try
    while DbiGetNextRecord(VCursor, dbiNoLock, @ValCheckDesc, nil) = 0 do
    begin
      NativeToAnsi(Locale, @ValCheckDesc.szPict, s);
      if s <> '' then
        FPxMasks.Values[FieldDefs[ValCheckDesc.iFldNum - 1].Name] := s;
    end;
  finally
    DbiCloseCursor(VCursor);
    if (FPxMasks.Count = 0) then ClearPxMasks;
  end;
end;

end.
